"""Implementation of XBAreaction class.

Peter Schubert, HHU Duesseldorf, May 2022
"""

import re
import sbmlxdf
from xbanalysis.utils.utils import get_species_stoic, get_local_params, expand_kineticlaw


class XbaReaction:

    def __init__(self, s_reaction, functions, compartments):
        self.id = s_reaction.name
        self.name = s_reaction.get('name', '')
        self.sboterm = s_reaction['sboterm']
        self.reaction_string = s_reaction['reactionString']
        self.reversible = s_reaction['reversible']
        self.products = get_species_stoic(s_reaction['products'])
        self.reactants = get_species_stoic(s_reaction['reactants'])
        self.modifiers = get_species_stoic(s_reaction['modifiers'])
        self.fbc_lb = s_reaction.get('fbcLb', float('nan'))
        self.fbc_ub = s_reaction.get('fbcUb', float('nan'))
        self.local_params = get_local_params(s_reaction)
        self.kinetic_law = s_reaction['kineticLaw'] if type(s_reaction['kineticLaw']) == str else ''
        # convert to numpy, replace local parameters with numerical values, inline lambda functions:
        expanded_kl = sbmlxdf.misc.mathml2numpy(self.kinetic_law)
        for pid, val in self.local_params.items():
            expanded_kl = re.sub(r'\b' + pid + r'\b', str(val), expanded_kl)
        self.expanded_kl = expand_kineticlaw(expanded_kl, functions)
        for cid in compartments.keys():
            if re.search(r'\b'+cid+r'\b', self.expanded_kl) is not None:
                self.compartment = cid
        self.ps_product = ''
        self.enzyme = ''

    def set_enzyme(self, species):
        for sid in self.modifiers:
            if species[sid].sboterm in ('SBO:0000252', 'SBO:0000250'):
                self.enzyme = sid
                break

    def add_species_usage(self, species):
        # for metabolic reactions add to the species consumed/produced the metabolic reaction id
        # !!! we could also enter the Reaction Object references instead
        if self.sboterm in ('SBO:0000167', 'SBO:0000179', 'SBO:0000182'):
            for sid, val in self.products.items():
                species[sid].add_m_reaction(self.id, val)
            for sid, val in self.reactants.items():
                species[sid].add_m_reaction(self.id, -val)
        # protein degradation reactions
        if self.sboterm == 'SBO:0000179':
            for sid, val in self.reactants.items():
                if species[sid].sboterm in ('SBO:0000252', 'SBO:0000250'):
                    species[sid].add_p_degrad_rid(self.id, -val)
        # protein transitions
        if self.sboterm == 'SBO:0000182':
            for sid, val in self.reactants.items():
                if species[sid].sboterm in ('SBO:0000252', 'SBO:0000250'):
                    species[sid].add_p_transition_rid(self.id, -val)
            for sid, val in self.products.items():
                if species[sid].sboterm in ('SBO:0000252', 'SBO:0000250'):
                    species[sid].add_p_transition_rid(self.id, val)

        # protein synthesis reactions
        if self.sboterm in ('SBO:0000589', 'SBO:0000205'):
            for sid, val in self.products.items():
                if species[sid].sboterm in ('SBO:0000252', 'SBO:0000250'):
                    self.ps_product = sid
                    species[sid].set_ps_rid(self.id)
                else:
                    species[sid].add_ps_reaction(self.id, val)
            for sid, val in self.reactants.items():
                species[sid].add_ps_reaction(self.id, -val)
